#!/usr/bin/env/python3

# Copyright (c) Facebook, Inc. and its affiliates.
# All rights reserved.
#
# This source code is licensed under the BSD-style license found in the
# LICENSE file in the root directory of this source tree.

import abc
from typing import Any, Callable, Dict, Tuple


class RendezvousClosedException(Exception):
    """
    Raised when a rendezvous for the specified run_id is closed.
    This is used to signal completion to nodes that arrive late.
    """

    pass


class RendezvousTimeoutException(Exception):
    """
    Raised from ``RendezvousHandler.next_rendezvous()`` to signal that the
    rendezvous did not
    succeed within the allocated time. This is meant to be interpreted
    as a non-retryable type of failure.
    """

    pass


class RendezvousNonRetryableError(Exception):
    """
    Raised from any of the ``RendezvousHandler`` methods when a failure
    occured that should not be retried with the same worker process.
    """

    pass


class RendezvousHandler(abc.ABC):
    """
    Main rendezvous interface.

    .. note:: torchelastic users normally **do not** need to implement their
              own ``RendezvousHandler``. An implementation based on
              `etcd <https://etcd.io/>`__ is already provided, and is recommended
              for most users, provided they can deploy it in their environment.

    .. warning:: torchelastic is currently considered experimental,
                 so the APIs may change!
    """

    @abc.abstractmethod
    def next_rendezvous(
        self,
        # pyre-ignore[11]: Annotation `Store` is not defined as a type.
        # pyre-ignore[10]: Name `torch` is used but not defined.
    ) -> Tuple["torch.distributed.Store", int, int]:  # noqa: F821
        """
        Main entry-point into the rendezvous barrier.
        Blocks until the rendezvous is complete (and the current
        process is included in the formed worker group), or a timeout occurs, or
        rendezvous was marked closed.

        Returns: a tuple of (``c10d Store``, ``rank``, ``world size``)

        Raises:
            RendezvousClosedException - if rendezvous for the current
               job is closed.
            RendezvousTimeoutException - on timeout
        """
        pass

    @abc.abstractmethod
    def is_closed(self) -> bool:
        """
        Checks whether rendezvous for current job has been closed,
        which means all future attempts to re-rendezvous (within same job) will
        fail.

        .. note:: ``is_closed`` and ``set_closed`` have semantics of eventual
                  propagation, and should not be used for synchronization.
                  The intention here is that if at least one worker decides
                  the job is finished, it will close the rendezvous, and
                  other workers will soon observe this and stop
                  training/rendezvous-ing as well.
        """
        pass

    @abc.abstractmethod
    def set_closed(self):
        """
        Used to mark the rendezvous (for current job) as closed.
        """
        pass

    @abc.abstractmethod
    def num_nodes_waiting(self) -> int:
        """
        Returns number of workers who *arrived late* at
        the rendezvous barrier, hence weren’t included in the current worker
        group.

        Callers should periodically call this method to check whether
        new members are waiting to join the job and if so admit them by
        calling ``next_rendezvous()`` (re-rendezvous).
        """
        pass

    @abc.abstractmethod
    def get_run_id(self) -> str:
        """
        Returns the run_id of this rendezvous handler. The run_id is a user-defined
        id that uniquely identifies an instance of a distributed application.
        It typically maps to a job id and is used to allow workers to join the
        correct distributed application.
        """
        pass

    def shutdown(self) -> bool:
        """
        Closes all resources that were open for rendezvous run.

        Usage:

        ::

         def main():
             rdzv_handler = ...
             try:
               rank, world_size, store = rdzv_handler.next_rendezvous()
             finally:
               rdzv_handler.shutdown()
        """
        pass


class RendezvousParameters(object):
    """
    Data object holding necessary and sufficient configuration parameters to
    construct a ``RendezvousHandler`` specified by the ``rdzv_backend``.
    """

    __slots__ = ("backend", "endpoint", "run_id", "min_nodes", "max_nodes", "configs")

    def __init__(
        self,
        backend: str,
        endpoint: str,
        run_id: str,
        min_nodes: int,
        max_nodes: int,
        **kwargs,
    ):
        """
        backend: The rdzv_backend that is used to register rendezvous, e.g. etcd
        endpoint: The rdzv_endpoint of the rendezvous. Usually it is a string
                    in the format host:port
        run_id: The unique job id that is used in rendezvous.
        min_nodes: The min number of nodes required to complete rendezvous.
        max_nodes: The max amount of nodes that are allowed to join the rendezvous.
        **kwargs: Additional configurations for the particular rdzv backend as key, value pairs
        """
        self.backend = backend
        self.endpoint = endpoint
        self.run_id = run_id
        self.min_nodes = min_nodes
        self.max_nodes = max_nodes

        self.configs = {}
        for (key, val) in kwargs.items():
            self.configs[key] = val

        assert 0 < min_nodes <= max_nodes
        assert backend is not None

    def get(self, config_key: str, default_value: Any = None) -> Any:
        """
        Returns the the config value from the configs map
        if one exits or the ``default_value``. Checks for ``None`` values,
        so if the config key exists in ``configs``, but maps to ``None``, then
        the ``default_value`` is returned.

        If the default_value is not specified (or is ``None``) then this method
        interprets the config key as a required config and will raise a ``KeyError``
        if the config key is either not found or maps to ``None``
        """
        if config_key not in self.configs:
            if default_value is None:
                raise KeyError(
                    f"required config: {config_key} not found, and a default was not provided"
                )
            else:
                return default_value

        val = self.configs[config_key]
        if val is None:
            if default_value is not None:
                return default_value
            else:
                raise KeyError(
                    f"required config: {config_key} maps to None, and a default was not provided"
                )
        else:
            return val


class RendezvousHandlerFactory:
    def __init__(self):
        self._factory_method_registry: Dict[
            str, Callable[[RendezvousParameters], RendezvousHandler]
        ] = {}

    def register(
        self,
        rdzv_backend: str,
        factory_method: Callable[[RendezvousParameters], RendezvousHandler],
    ):
        if rdzv_backend in self._factory_method_registry:
            fn = self._factory_method_registry[rdzv_backend]
            raise ValueError(
                f"cannot double register rdzv_backend: {rdzv_backend} "
                f"to {factory_method.__module__}.{factory_method.__name__},"
                f" already registered with: {fn.__module__}.{fn.__name__}"
            )

        self._factory_method_registry[rdzv_backend] = factory_method

    def create_rdzv_handler(self, rdzv_params: RendezvousParameters):
        backend = rdzv_params.backend
        if backend not in self._factory_method_registry:
            raise ValueError(
                f"no factory method found for rdzv_backend: {backend},"
                f" did you forget to call {self.register.__name__}?"
            )

        return self._factory_method_registry[backend](rdzv_params)



#!/usr/bin/env python3

# Copyright (c) Facebook, Inc. and its affiliates.
# All rights reserved.
#
# This source code is licensed under the BSD-style license found in the
# LICENSE file in the root directory of this source tree.


import abc
import os
import subprocess
import sys
import time
from typing import Dict, List, Optional

from torchelastic.tsm.driver.api import (
    Application,
    AppState,
    DescribeAppResponse,
    RunMode,
    Scheduler,
    is_terminal,
    macros,
)
from torchelastic.utils.logging import get_logger


log = get_logger()


class ImageFetcher(abc.ABC):
    """
    Downloads and sets up an image onto the localhost. This is only needed for
    ``LocalhostScheduler`` since typically real schedulers will do this
    on-behalf of the user.
    """

    @abc.abstractmethod
    def fetch(self, image: str) -> str:
        """
        Pulls the given image and returns a path to the pulled image on
        the local host.
        """
        raise NotImplementedError()


class LocalDirectoryImageFetcher(ImageFetcher):
    """
    Interprets the image name as the path to a directory on
    local host. Does not "fetch" (e.g. download) anything. Used in conjunction
    with ``LocalScheduler`` to run local binaries.

    The image name must be an absolute path and must exist.

    Example:

    1. ``fetch(Image(name="/tmp/foobar"))`` returns ``/tmp/foobar``
    2. ``fetch(Image(name="foobar"))`` raises ``ValueError``
    2. ``fetch(Image(name="/tmp/dir/that/does/not_exist"))`` raises ``ValueError``
    """

    def fetch(self, image: str) -> str:
        """
        Raises:
            ValueError - if the image name is not an absolute dir
                         and if it does not exist or is not a directory

        """
        if not os.path.isabs(image):
            raise ValueError(
                f"Invalid image name: {image}, image name must be an absolute path"
            )

        if not os.path.isdir(image):
            raise ValueError(
                f"Invalid image name: {image}, does not exist or is not a directory"
            )

        return image


# aliases to make clear what the mappings are
AppId = str
AppName = str
RoleName = str


class _LocalApplication:
    """
    Container object used by ``LocalhostScheduler`` to group the pids that
    form an application. Each replica of a role in the application is a
    process and has a pid.
    """

    def __init__(self, name: str):
        self.name = name
        # role name -> proc_1, proc_2, ... (proc for each replica)
        self.role_procs: Dict[RoleName, List[subprocess.Popen]] = {}
        self.state: AppState = AppState.PENDING
        # time (in seconds since epoch) when the last set_state method() was called
        self.last_updated: float = -1
        self.run_mode: RunMode = RunMode.MANAGED

    def add_process(self, role_name: str, proc: subprocess.Popen) -> None:
        procs = self.role_procs.setdefault(role_name, [])
        procs.append(proc)

    def set_state(self, state: AppState):
        self.last_updated = time.time()
        self.state = state

    def set_run_mode(self, mode: RunMode):
        self.run_mode = mode

    def __repr__(self):
        role_to_pid = {}
        for (role_name, procs) in self.role_procs.items():
            pids = role_to_pid.setdefault(role_name, [])
            for p in procs:
                pids.append(p.pid)

        return f"{{name:{self.name}, state:{self.state}, mode:{self.run_mode}, pid_map:{role_to_pid}}}"


class LocalScheduler(Scheduler):
    """
    Schedules on localhost. Containers are modeled as processes and
    certain properties of the container that are either not relevant
    or that cannot be enforced for localhost
    runs are ignored. Properties that are ignored:

    1. Resource requirements
    2. Container limit enforcements
    3. Restart policies
    4. Retry counts (no retries supported)

    ..note:: Use this scheduler sparingly since an application
             that runs successfully on a session backed by this
             scheduler may not work on an actual production cluster
             using a different scheduler.
    """

    def __init__(self, image_fetcher: ImageFetcher, cache_size: int = 100):
        # for each app name keeps an incrementing id, to support running
        # multiple instances of the same app
        self._ids: Dict[AppName, int] = {}
        # TODO T72035686 replace dict with a proper LRUCache data structure
        # and don't forget to remove the app name from "_ids" when evicting finished apps=
        self._apps: Dict[AppId, _LocalApplication] = {}
        self._image_fetcher = image_fetcher

        if cache_size <= 0:
            raise ValueError("cache size must be greater than zero")
        self._cache_size = cache_size

    def _evict_lru(self) -> bool:
        """
        Evicts one least recently used element from the apps cache. LRU is defined as
        the oldest app in a terminal state (e.g. oldest finished app).

        Returns:
            ``True`` if an entry was evicted, ``False`` if no entries could be evicted
            (e.g. all apps are running)
        """
        lru_time = sys.maxsize
        lru_app_id = None
        for (app_id, app) in self._apps.items():
            if is_terminal(app.state):
                if app.last_updated <= lru_time:
                    lru_app_id = app_id

        if lru_app_id:
            # evict LRU finished app from the apps cache
            # do not remove the app name from the ids map so that the ids
            # remain unique throughout the lifespan of this scheduler
            # for example if cache size == 1
            #     app_id1 = submit(app)
            #     app_id2 = submit(app) # app_id1 was evicted here
            #     app_id1 == "app.name_0"
            #     app_id2 == "app.name_1"
            del self._apps[lru_app_id]

            log.debug(f"evicting app: {lru_app_id}, from local scheduler cache")
            return True
        else:
            log.debug(f"no apps evicted, all {len(self._apps)} apps are running")
            return False

    def submit(self, app: Application, mode: RunMode) -> str:
        if len(self._apps) == self._cache_size:
            if not self._evict_lru():
                raise IndexError(
                    f"App cache size ({self._cache_size}) exceeded. Increase the cache size"
                )

        id = self._ids.setdefault(app.name, -1) + 1
        self._ids[app.name] = id
        app_id = f"{app.name}_{id}"

        assert (
            app_id not in self._apps
        ), "no app_id collisons expected since incremental integer suffix is used"

        local_app = _LocalApplication(app.name)
        local_app.set_run_mode(mode)

        for role in app.roles:
            container = role.container
            assert (
                container
            ), "all roles in a submitted app must have container association"

            img_root = self._image_fetcher.fetch(container.image)
            cmd = os.path.join(img_root, role.entrypoint)
            for replica_id in range(role.num_replicas):
                args = [cmd] + macros.substitute(
                    role.args, img_root, app_id, str(replica_id)
                )
                log.info(f"Running {args} with env: {role.env}")
                proc = subprocess.Popen(args, env=role.env)
                local_app.add_process(role.name, proc)

        self._apps[app_id] = local_app
        return app_id

    def _failed(self, p: subprocess.Popen):
        if self._is_alive(p):
            return False
        else:
            return p.returncode != 0

    def describe(self, app_id: str) -> Optional[DescribeAppResponse]:
        if app_id not in self._apps:
            return None

        local_app = self._apps[app_id]

        # check if the app has been known to have finished
        if is_terminal(local_app.state):
            state = local_app.state
        else:
            running = False
            failed = False
            for (_, procs) in local_app.role_procs.items():
                for p in procs:
                    running |= self._is_alive(p)
                    failed |= self._failed(p)

            if running:
                state = AppState.RUNNING
            elif failed:
                state = AppState.FAILED
                self._terminate(local_app)  # terminate danglers
            else:
                state = AppState.SUCCEEDED
            local_app.set_state(state)

        resp = DescribeAppResponse()
        resp.app_id = app_id
        resp.state = state
        resp.num_restarts = 0
        return resp

    def _is_alive(self, p: subprocess.Popen):
        return p.poll() is None

    def _terminate(self, app: _LocalApplication):
        for (_, procs) in app.role_procs.items():
            for p in procs:
                # safe to call terminate on a process that already died
                p.terminate()
                p.wait()

    def _cancel_existing(self, app_id: str) -> None:
        # can assume app_id exists
        local_app = self._apps[app_id]
        self._terminate(local_app)
        local_app.state = AppState.CANCELLED

    def wait(self, app_id: str) -> Optional[DescribeAppResponse]:
        """
        Waits for the app to finish or raise TimeoutError upon timeout (in seconds).
        If no timeout is specified waits indefinitely.

        Returns:
            The last return value from ``describe()``
        """

        expiry = sys.maxsize
        interval = 1

        while expiry > time.time():
            desc = self.describe(app_id)

            if desc is None:
                return None
            elif is_terminal(desc.state):
                return desc

            time.sleep(interval)

        raise TimeoutError(f"timed out waiting for app: {app_id} to finish")

    def __del__(self):
        # terminate all MANAGED apps
        for (app_id, app) in self._apps.items():
            if app.run_mode == RunMode.MANAGED:
                log.info(f"Terminating managed app: {app_id}")
                self._cancel_existing(app_id)


def create_scheduler(**kwargs) -> LocalScheduler:
    image_fetcher = LocalDirectoryImageFetcher()
    cache_size = kwargs.get("cache_size", 100)
    return LocalScheduler(image_fetcher, cache_size=cache_size)



#!/usr/bin/env python3

# Copyright (c) Facebook, Inc. and its affiliates.
# All rights reserved.
#
# This source code is licensed under the BSD-style license found in the
# LICENSE file in the root directory of this source tree.

# Follows PEP-0440 version scheme guidelines
# https://www.python.org/dev/peps/pep-0440/#version-scheme
#
# Examples:
# 0.1.0.devN # Developmental release
# 0.1.0aN  # Alpha release
# 0.1.0bN  # Beta release
# 0.1.0rcN  # Release Candidate
# 0.1.0  # Final release
__version__ = "0.2.1"



#!/usr/bin/env python3

# Copyright (c) Facebook, Inc. and its affiliates.
# All rights reserved.
#
# This source code is licensed under the BSD-style license found in the
# LICENSE file in the root directory of this source tree.

import inspect
import logging
import os
from typing import Optional


def get_logger(name: Optional[str] = None):
    r"""
    Util function to set up a simple logger that writes
    into stderr. The loglevel is fetched from the LOGLEVEL
    env. variable or INFO as default. The function will use the
    module name of the caller if no name is provided.

    Arguments:
        name (str): Name of the logger. If no name provided, the name will
        be derived from the call stack.
    """

    if name is None:
        try:
            # Derive the name of the caller. Since the module name derivation
            # is a function we use the depth=2 to find the caller.
            name = _derive_module_name(depth=2)
        except Exception as e:
            default_log = _setup_logger()
            default_log.warn(
                f"Error while setting up logger. Will be using default logger. Got exception: {e}"
            )

    log = _setup_logger(name)
    return log


def _setup_logger(name: Optional[str] = None):
    log = logging.getLogger(name)
    if len(log.handlers) == 0:
        # Add default handler that writes messages to stderr
        log.addHandler(logging.StreamHandler())
    log.setLevel(os.environ.get("LOGLEVEL", "INFO"))
    return log


def _derive_module_name(depth: int = 1) -> str:
    r"""
    Derives the name of the caller module from the stack frames.

    Arguments:
        depth (int): The position of the frame in the stack.
    """
    stack = inspect.stack()
    assert depth < len(stack)
    node = stack[depth]
    # Each element of the stack is an array of elements, where the
    # first element should alway be a fame object.
    frame = node[0]
    module = inspect.getmodule(frame)
    if module is None:
        raise ValueError(f"Frame {frame} at depth {depth} does not have module.")
    return module.__name__



#!/usr/bin/env python3

# Copyright (c) Facebook, Inc. and its affiliates.
# All rights reserved.
#
# This source code is licensed under the BSD-style license found in the
# LICENSE file in the root directory of this source tree.
# Multiprocessing error-reporting module

import logging
import os
from typing import Any, Callable, Optional, Tuple

from torchelastic.multiprocessing.error_reporter.handlers import get_error_handler


log: logging.Logger = logging.getLogger(__name__)


def exec_fn(user_funct: Callable, args: Tuple = ()) -> Any:
    log.warning("This is an experimental API and is going to be removed in future")
    error_handler = get_error_handler()
    error_handler.configure()
    return user_funct(*args)


def get_error(error_process_pid: int) -> Optional[str]:
    """
    Retrieves an error(if any) that occurred on the child process.
    If no exception occurred, the function will return None
    """
    log.warning("This is an experimental API and is going to be removed in future")
    error_handler = get_error_handler()
    return error_handler.construct_error_message(error_process_pid)


def _configure_process_handler() -> None:
    log.warning("This is an experimental API and is going to be removed in future")
    error_handler = get_error_handler()
    error_handler.configure()


def record(func):
    """
    Decorator function that is invoked on the starting process function.
    The decorator will invoke signal registerring mechanism that will allow
    to propagate any signal termination related errors to the parent process.

    Example

    ::

        from torchelastic.multiprocessing.error_reporter import record

        @record
        def main():
            ...

        if __name__=="__main__":
            main()

    """

    log.warning("This is an experimental API and is going to be removed in future")

    def wrap():
        try:
            _configure_process_handler()
            func()
        except Exception as e:
            error_handler = get_error_handler()
            error_handler.record_exception(e)
            raise

    return wrap



#!/usr/bin/env python3

# Copyright (c) Facebook, Inc. and its affiliates.
# All rights reserved.
#
# This source code is licensed under the BSD-style license found in the
# LICENSE file in the root directory of this source tree.

import abc
import time
from typing import Any, Dict


class Event:
    """
    The class represents the generic event that occurs during the torchelastic
    job execution. The event can be any kind of meaningful action.

    Args:
        name: event name.
        type: type of event, determines what handler will process the event.
        metadata: additional data that is associated with the event.
        timestamp: timestamp in milliseconds when event occured.
    """

    __slots__ = ["name", "type", "metadata", "timestamp"]

    def __init__(
        self,
        name: str,
        type: str = "torchelastic",
        # Cannot use optional due to https://github.com/facebook/pyre-check/issues/197
        # pyre-fixme[9]: metadata has type `Dict[str, Any]`; used as `None`.
        metadata: Dict[str, Any] = None,
        timestamp: int = 0,
    ):
        self.name = name
        self.type = type
        self.metadata = metadata
        self.timestamp = timestamp


class EventHandler(abc.ABC):
    """
    Event handler interface is responsible for capturing and recording events
    to the different destinations, e.g. stdout, file, etc.
    """

    @abc.abstractmethod
    def record(self, event: Event):
        raise NotImplementedError()


class ConsoleEventHandler(EventHandler):
    """
    The event handler that prints the captured events to the console.
    """

    def record(self, event: Event):
        print(f"[{event.timestamp}] {event.type}:{event.name}")


class NullEventHandler(EventHandler):
    """
    The default event handler that is a no-op.
    """

    def record(self, event: Event):
        pass


_events_map: Dict[str, EventHandler] = {}
_default_event_handler: EventHandler = NullEventHandler()


def _getEventHandler(event_type: str):
    if event_type in _events_map:
        return _events_map[event_type]
    return _default_event_handler


def configure(handler: EventHandler, event_type: str = ""):
    r"""
    Sets the relation between event_type and handler. After the relation is set
    all events that have type `event_type` will be processed by the handler.

    Args:
        handler: Event handler that will process the events.
        event_type: Type of event.
    """
    if not event_type:
        global _default_event_handler
        _default_event_handler = handler
    else:
        _events_map[event_type] = handler


def record_event(event: Event):
    r"""
    Records the event to the destination configured by the ``event.type`` parameter.

    Args:
        event (Event): Event to be recorded
    """
    if event.metadata is None:
        event.metadata = {}
    _getEventHandler(event.type).record(event)


def record(
    event_name: str,
    event_type: str = "torchelastic",
    # Cannot use optional due to https://github.com/facebook/pyre-check/issues/197
    # pyre-fixme[9]: metadata has type `Dict[str, Any]`; used as `None`.
    metadata: Dict[str, Any] = None,
):
    r"""
    Constructs and  records the event to the destination
    configured by the event.type parameter.

    Args:
        event_name: The name of the event
        event_type: The type of the event. Is used to determine what handler
            will process the event
        metadata: Arbitrary data that is associated with the event.
    """

    if metadata is None:
        metadata = {}
    event = Event(
        event_name, event_type, metadata, timestamp=int(round(time.time() * 1000))
    )
    record_event(event)



#!/usr/bin/env python3

# Copyright (c) Facebook, Inc. and its affiliates.
# All rights reserved.
#
# This source code is licensed under the BSD-style license found in the
# LICENSE file in the root directory of this source tree.

r"""
This module provides similar functionality as ``torch.distributed.launch``,
with the following additional functionalities:

1. Worker failures are handled gracefully by restarting all workers.

2. Worker ``RANK`` and ``WORLD_SIZE`` are assigned automatically.

3. Number of nodes is allowed to change between min and max sizes (elasticity).

**Usage:**

1. Single-node multi-worker (with sidecar etcd server)

::

    >>> python -m torchelastic.distributed.launch
        --standalone
        --nnodes=1
        --nproc_per_node=$NUM_TRAINERS
        YOUR_TRAINING_SCRIPT.py (--arg1 ... train script args...)

2. Fault tolerant (fixed sized number of workers, no elasticity).:

::

    >>> python -m torchelastic.distributed.launch
        --nnodes=$NUM_NODES
        --nproc_per_node=$NUM_TRAINERS
        --rdzv_id=$JOB_ID
        --rdzv_backend=etcd
        --rdzv_endpoint=$ETCD_HOST:$ETCD_PORT
        YOUR_TRAINING_SCRIPT.py (--arg1 ... train script args...)

3. Elastic (``min=1``, ``max=4``):

::

    >>> python -m torchelastic.distributed.launch
        --nnodes=1:4
        --nproc_per_node=$NUM_TRAINERS
        --rdzv_id=$JOB_ID
        --rdzv_backend=etcd
        --rdzv_endpoint=$ETCD_HOST:$ETCD_PORT
        YOUR_TRAINING_SCRIPT.py (--arg1 ... train script args...)

**Note on rendezvous backend**:

For multi-node training you need to specify:

1. ``--rdzv_id``: a unique job id (shared by all nodes participating in the job)
2. ``--rdzv_backend``: an implementation of ``torchelastic.rendevous.RendezvousHandler``
3. ``--rdzv_endpoint``: ``host:port``-style endpoint where the rdzv backend is running.

Currently only ``etcd`` rdzv backend is supported out of the box.
To use ``etcd``, setup an etcd server with the ``v2`` api enabled
(e.g. ``--enable-v2``).

.. warning:: ``EtcdRendezvous`` uses etcd api v2. You MUST enable the v2
             api on the etcd server. Our tests use etcd v3.4.3.

**Definitions:**

1. ``Node`` - Physical instance or container.
    Maps to the unit that the job manager works with.

2. ``Worker`` - A worker in the context of distributed training.

3. ``Worker Group`` - Workers that execute the same function (e.g. trainers)

4. ``Local Worker Group`` - Subset of the workers in the
    worker group running on the same Node

5. ``RANK`` - rank of the worker within a worker group.

6. ``WORLD_SIZE`` - total number of workers in a worker group.

7. ``LOCAL_RANK`` - rank of the worker within a local worker group

8. ``LOCAL_WORLD_SIZE`` - size of the local worker group

9. ``rdzv_id`` - user defined id that uniquely identifies the worker group
      for a job. This id is used by each node to join as a member of a particular
      worker group.

9. ``rdzv_backend`` - the backend store of rendezvous (e.g. etcd). This is
    typically a strongly consistent key-value store.

10. ``rdzv_endpoint`` - rdzv backend server endpoint in ``host:port`` format.

A ``Node`` runs ``LOCAL_WORLD_SIZE`` workers which comprise a ``LocalWorkerGroup``.
The union of all ``LocalWorkerGroups`` in the nodes in the job comprise the
``WorkerGroup``.

**Environment Variables:**

The following environment variables are made available to you in your
script:

1. ``LOCAL_RANK`` -  local rank

2. ``RANK`` -  global rank

3. ``GROUP_RANK`` - rank of the worker group. A number between 0 - ``max_nnodes``.
        When running a single worker group per node, this is the rank of the node.

4. ``ROLE_RANK`` -  the rank of the worker across all the workers tha have the same
        role. The role of the worker is specified in the ``WorkerSpec``.

5. ``LOCAL_WORLD_SIZE`` - local world size (e.g. number of workers running locally).
       Equal to ``--nproc_per_node`` specified on ``torchelastic.distributed.launch``.

6. ``WORLD_SIZE`` - world size (total number of workers in the job).

7. ``ROLE_WORLD_SIZE`` - the total number of workers that was launched with the same
        role specified in ``WorkerSpec``.

8. ``MASTER_ADDR`` - fqdn of the host that is running worker with rank 0.
   Used to initialize torch distributed backend.

9. ``MASTER_PORT`` - port on the ``MASTER_ADDR`` that can be used to
   host the tcp ``c10d`` store.

10. ``TORCHELASTIC_RESTART_COUNT`` - number of worker group restarts so far.

11. ``TORCHELASTIC_MAX_RESTARTS`` - configured max number of restarts.

12. ``TORCHELASTIC_RUN_ID`` - equal to rdzv run_id (e.g. unique job id).

**Deployment:**

1. Start the rdzv backend server and get the endpoint
   (to be passed as ``--rdzv_endpoint`` to the launcher script)

2. Single-node multi-worker - start the launcher on the host to start
   the agent process which creates and monitors a local worker group.

3. Multi-node multi-worker - Start the launcher with the same arguments
   on all the nodes participating in training.

When using a job/cluster manager the entry point command to the multi-node
job is invoking this launcher.

**Failure Modes:**

1. Worker failure - For a training job with ``n`` workers, if ``k <= n`` workers fail
   all workers are stopped and restarted up to ``max_restarts``.

2. Agent failure - An agent failure results in local worker group failure,
   it is up to the job manager to fail the entire job (gang semantics) or attempt
   to replace the node. Both behaviors are supported by the agent.

3. Node failure - Same as agent failure.

**Membership Changes:**

1. Node departure (scale-down) - agent is notified of the departure,
   all existing workers are stopped, a new ``Worker Group`` is formed and all
   workers are started with a new ``RANK`` and ``WORLD_SIZE``.

2. Node arrival (scale-up) - the new node is admitted to the job,
   all existing workers are stopped, a new ``Worker Group`` is formed and all
   workers are started with a new ``RANK`` and ``WORLD_SIZE``.


**Important Notices:**

1. All the items in the important notices section of ``torch.distributed.launch``
   apply to this module as well

2. The environment variables necessary to initialize a torch process group
   are provided to you by this module, no need for you to pass ``RANK`` manually.
   To initialize a process group in your training script, simply run

::

 >>> import torch.distributed as dist
 >>> dist.init_process_group(backend="gloo|nccl")

3. On failures or membership changes ALL surviving workers are killed
   immediately. Make sure to checkpoint your progress. The frequency of
   checkpoints should depend on your job's tolerance for lost work.

4. This module only supports homogeneous ``LOCAL_WORLD_SIZE``. That is,
   it is assumed that all nodes run the same number of local workers (per role).

5. ``RANK`` is NOT stable. Between restarts, the local workers on a node
   can be assgined a different range of ranks than before. NEVER hard code
   any assumptions about the stable-ness of ranks or some correlation between
   ``RANK`` and ``LOCAL_RANK``.

6. When using elasticity (``min_size != max_size``) DO NOT hard code
   assumptions about ``WORLD_SIZE`` as the world size can change as
   nodes are allowed to leave and join.

7. It is recommended your script have the following structure

::

  def main():
    load_checkpoint(checkpoint_path)
    initialize()
    train()

  def train():
    for batch in iter(dataset):
      train_step(batch)

      if should_checkpoint:
        save_checkpoint(checkpoint_path)
"""
import logging
import os
import signal
import subprocess
import sys
import uuid
from argparse import REMAINDER, ArgumentParser

import torch
import torchelastic.rendezvous.registry as rdzv_registry
from torchelastic import metrics
from torchelastic.agent.server.api import WorkerGroupFailureException, WorkerSpec
from torchelastic.agent.server.local_elastic_agent import LocalElasticAgent
from torchelastic.distributed.argparse_util import check_env, env
from torchelastic.rendezvous import RendezvousParameters
from torchelastic.rendezvous.etcd_server import EtcdServer
from torchelastic.utils.logging import get_logger


log = get_logger()


def parse_args(args):
    """
    Helper function parsing the command line options.
    """

    parser = ArgumentParser(description="torchelastic elastic training launcher")

    # Arguments for the launch helper
    # worker/node size related arguments
    parser.add_argument(
        "--nnodes",
        action=env,
        type=str,
        default="1:1",
        help="number of nodes or MIN_NODES:MAX_NODES",
    )
    parser.add_argument(
        "--nproc_per_node",
        action=env,
        type=str,
        default="auto",
        help="number of workers per node, supported values: [auto, cpu, gpu, int]",
    )

    # rendezvous related arguments
    parser.add_argument(
        "--rdzv_backend",
        action=env,
        type=str,
        default="etcd",
        help="rendezvous backend",
    )
    parser.add_argument(
        "--rdzv_endpoint",
        action=env,
        type=str,
        default="",
        help="rendezvous backend server host:port",
    )
    parser.add_argument("--rdzv_id", action=env, type=str, help="user defined group id")
    parser.add_argument(
        "--rdzv_conf",
        action=env,
        type=str,
        default="",
        help="additional rdzv configuration (conf1=v1,conf2=v2,...)",
    )

    # sidecar embed rdzv backend that defaults to etcd
    parser.add_argument(
        "--standalone",
        action=check_env,
        help="starts a local, standalone rdzv backend that is represented by"
        " etcd server on a random free port"
        "using the etcd binary specified in TORCHELASTIC_ETCD_BINARY_PATH"
        " env var or the one found in PATH."
        " Useful when launching single-node, multi-worker job."
        " If specified --rdzv_backend, --rdzv_endpoint, --rdzv_id"
        " are autoassigned, any explicitly set values are ignored",
    )

    # user-code launch related arguments
    parser.add_argument(
        "--max_restarts",
        action=env,
        type=int,
        default=3,
        help="max number of worker group restarts before failing",
    )
    parser.add_argument(
        "--monitor_interval",
        action=env,
        type=float,
        default=5,
        help="interval (in seconds) to monitor the state of workers",
    )
    parser.add_argument(
        "--start_method",
        action=env,
        type=str,
        default="spawn",
        choices=["spawn", "fork", "forkserver"],
        help="multiprocessing start_method to use when creating workers",
    )
    parser.add_argument(
        "--role",
        action=env,
        type=str,
        default="default",
        help="user-defined role for the workers",
    )
    parser.add_argument(
        "-m",
        "--module",
        action=check_env,
        help="Changes each process to interpret the launch script "
        "as a python module, executing with the same behavior as"
        "'python -m'.",
    )
    parser.add_argument(
        "--no_python",
        action=check_env,
        help='Do not prepend the training script with "python" - just exec '
        "it directly. Useful when the script is not a Python script.",
    )

    # positional
    parser.add_argument(
        "training_script",
        type=str,
        help="The full path to the single GPU training "
        "program/script to be launched in parallel, "
        "followed by all the arguments for the "
        "training script",
    )

    # rest from the training program
    parser.add_argument("training_script_args", nargs=REMAINDER)
    return parser.parse_args(args)


def parse_min_max_nnodes(nnodes: str):
    arr = nnodes.split(":")

    if len(arr) == 1:
        min_nodes = max_nodes = int(arr[0])
    elif len(arr) == 2:
        min_nodes = int(arr[0])
        max_nodes = int(arr[1])
    else:
        raise RuntimeError(f'nnodes={nnodes} is not in "MIN:MAX" format')

    return min_nodes, max_nodes


def wrapper_fn(omp_num_threads, cmd):
    # TODO get rid of this wrapper_fn when we switch over to torchelastic.multiprocessing
    # the agent uses multiprocessing.spawn to create nproc_per_node
    # instances of fn, and hence expects fn to be a callable
    # since this launcher deals with user python scripts and executables
    # we wrap the script/executable with this function which Popens
    # the wrapped script/executable. This implies that for each
    # worker we create two processes (wrapper_fn and fn).
    # the process tree looks like the following:
    #
    # [launcher/agent]
    #               |-- [wrapper_fn_0]
    #               |               |-- [fn_0]
    #               |-- [wrapper_fn_1]
    #               |               |-- [fn_1]
    #               |      ...
    #               |      ...
    #               |-- [wrapper_fn_k]
    #               |               |-- [fn_k]
    #

    # set PyTorch distributed related environmental variables
    if omp_num_threads is not None:
        os.environ["OMP_NUM_THREADS"] = str(omp_num_threads)

    process = subprocess.Popen(cmd)

    # since we wrap the script process with this function (which runs as a
    # subprocess of the agent) when the agent terminates this function
    # due to some exception or membership change event we want the script
    # to also get killed. If we do not register this exit handler
    # the script process will get re-parented to the parent of this function
    # (agent process) and we will end up with multiple copies of the script
    # this should all go away with D20613415
    def kill_script_pid(signum, frame):
        process.terminate()

    signal.signal(signal.SIGTERM, kill_script_pid)

    process.wait()
    if process.returncode != 0:
        raise subprocess.CalledProcessError(returncode=process.returncode, cmd=cmd)


def determine_local_world_size(nproc_per_node: str):
    try:
        logging.info(f"Using nproc_per_node={nproc_per_node}.")
        return int(nproc_per_node)
    except ValueError:
        if nproc_per_node == "cpu":
            num_proc = os.cpu_count()
            device_type = "cpu"
        elif nproc_per_node == "gpu":
            if not torch.cuda.is_available():
                raise ValueError("Cuda is not available.")
            device_type = "gpu"
            num_proc = torch.cuda.device_count()
        elif nproc_per_node == "auto":
            if torch.cuda.is_available():
                num_proc = torch.cuda.device_count()
                device_type = "gpu"
            else:
                num_proc = os.cpu_count()
                device_type = "cpu"
        else:
            raise ValueError(f"Unsupported nproc_per_node value: {nproc_per_node}")

        log.info(
            f"Using nproc_per_node={nproc_per_node},"
            f" seting to {num_proc} since the instance "
            f"has {os.cpu_count()} {device_type}"
        )
        return num_proc


def main(args=None):
    # If ``args`` not passed, defaults to ``sys.argv[:1]``
    args = parse_args(args)

    min_nodes, max_nodes = parse_min_max_nnodes(args.nnodes)
    assert 0 < min_nodes <= max_nodes
    assert args.max_restarts >= 0

    if args.standalone:
        etcd_server = EtcdServer()
        etcd_server.start()
        args.rdzv_backend = "etcd"
        args.rdzv_endpoint = etcd_server.get_endpoint()
        args.rdzv_id = str(uuid.uuid4())
        log.info(
            f"\n**************************************\n"
            f"Rendezvous info:\n"
            f"--rdzv_backend={args.rdzv_backend} "
            f"--rdzv_endpoint={args.rdzv_endpoint} "
            f"--rdzv_id={args.rdzv_id}\n"
            f"**************************************\n"
        )

    nproc_per_node = determine_local_world_size(args.nproc_per_node)
    omp_num_threads = None
    if "OMP_NUM_THREADS" not in os.environ and nproc_per_node > 1:
        omp_num_threads = 1
        print(
            f"*****************************************\n"
            f"Setting OMP_NUM_THREADS environment variable for each process to be "
            f"{omp_num_threads} in default, to avoid your system being overloaded, "
            f"please further tune the variable for optimal performance in "
            f"your application as needed. \n"
            f"*****************************************"
        )

    with_python = not args.no_python
    cmd = []
    if with_python:
        cmd = [sys.executable, "-u"]
        if args.module:
            cmd.append("-m")
    else:
        if args.module:
            raise ValueError(
                "Don't use both the '--no_python' flag"
                " and the '--module' flag at the same time."
            )

    cmd.append(args.training_script)
    cmd.extend(args.training_script_args)

    rdzv_parameters = RendezvousParameters(
        backend=args.rdzv_backend,
        endpoint=args.rdzv_endpoint,
        run_id=args.rdzv_id,
        min_nodes=min_nodes,
        max_nodes=max_nodes,
        **_parse_rdzv_conf(args.rdzv_conf),
    )

    rdzv_handler = rdzv_registry.get_rendezvous_handler(rdzv_parameters)

    try:
        spec = WorkerSpec(
            role=args.role,
            local_world_size=nproc_per_node,
            fn=wrapper_fn,
            args=(omp_num_threads, cmd),
            rdzv_handler=rdzv_handler,
            max_restarts=args.max_restarts,
            monitor_interval=args.monitor_interval,
        )
        metrics.initialize_metrics()
        elastic_agent = LocalElasticAgent(spec, start_method=args.start_method)
        elastic_agent.run(spec.role)
    except WorkerGroupFailureException as e:
        # user-code error; propagate the one from the smallest rank on this node
        excs = e.get_worker_exceptions()
        min_rank = min(excs.keys())
        raise excs[min_rank] from e
    except Exception:
        # this is an agent (platform) error
        raise
    finally:
        rdzv_handler.shutdown()
        if args.standalone:
            etcd_server.stop()


def _parse_rdzv_conf(conf_str: str):
    rdzv_conf = {}
    confs = conf_str.split(",")
    for kvs in confs:
        key = kvs.split("=")[0]
        value = kvs.split("=")[1:]
        rdzv_conf[key] = value
    return rdzv_conf


if __name__ == "__main__":
    logging.basicConfig(
        level=logging.INFO, format="[%(levelname)s] %(asctime)s %(module)s: %(message)s"
    )
    log.info(f"Running torchelastic.distributed.launch with args: {sys.argv}")
    main()



#!/usr/bin/env python3

# Copyright (c) Facebook, Inc. and its affiliates.
# All rights reserved.
#
# This source code is licensed under the BSD-style license found in the
# LICENSE file in the root directory of this source tree.

import os
from typing import Any, Dict

import torch.multiprocessing as mp
from torchelastic.agent.server.api import (
    MonitorResult,
    SimpleElasticAgent,
    Worker,
    WorkerGroup,
    WorkerSpec,
    WorkerState,
)
from torchelastic.metrics.api import prof
from torchelastic.utils.logging import get_logger


log = get_logger()


class _DistInfo:
    """
    Container for information required to create a torch process group.
    To be created on the agent's process and passed to the worker sub-process.
    Hence this object needs to be a pure data object with no state and
    preferably only primitive member variables
    """

    __slots__ = [
        "rank",
        "group_rank",
        "role_rank",
        "local_world_size",
        "role_world_size",
        "world_size",
        "master_addr",
        "master_port",
        "restart_count",
        "max_restarts",
        "run_id",
        "role_name",
    ]

    def __init__(
        self,
        rank: int,
        group_rank: int,
        role_rank: int,
        local_world_size: int,
        role_world_size: int,
        world_size: int,
        master_addr: str,
        master_port: int,
        restart_count: int,
        max_restarts: int,
        run_id: str,
        role_name: str,
    ):
        self.rank = rank
        self.group_rank = group_rank
        self.local_world_size = local_world_size
        self.role_rank = role_rank
        self.world_size = world_size
        self.role_world_size = role_world_size
        self.master_addr = master_addr
        self.master_port = master_port
        self.restart_count = restart_count
        self.max_restarts = max_restarts
        self.run_id = run_id
        self.role_name = role_name


def _wrap(local_rank, ret_val_queue, dist_infos, fn, args):
    import faulthandler

    try:
        faulthandler.enable(all_threads=True)
    except Exception as e:
        log.warn(
            "Unable to enable fault handler. Failure signals on worker process will not dump tracebacks",
            exc_info=e,
        )

    info = dist_infos[local_rank]
    os.environ["LOCAL_RANK"] = str(local_rank)
    os.environ["RANK"] = str(info.rank)
    os.environ["GROUP_RANK"] = str(info.group_rank)
    os.environ["ROLE_RANK"] = str(info.role_rank)
    os.environ["ROLE_NAME"] = info.role_name
    os.environ["LOCAL_WORLD_SIZE"] = str(info.local_world_size)
    os.environ["WORLD_SIZE"] = str(info.world_size)
    os.environ["ROLE_WORLD_SIZE"] = str(info.role_world_size)
    os.environ["MASTER_ADDR"] = info.master_addr
    os.environ["MASTER_PORT"] = str(info.master_port)
    os.environ["TORCHELASTIC_RESTART_COUNT"] = str(info.restart_count)
    os.environ["TORCHELASTIC_MAX_RESTARTS"] = str(info.max_restarts)
    os.environ["TORCHELASTIC_RUN_ID"] = info.run_id
    ret = fn(*args)
    ret_val_queue.put((info.rank, ret))


class LocalElasticAgent(SimpleElasticAgent):
    """
    An implementation of :py:class:`torchelastic.agent.server.ElasticAgent`
    that handles host-local workers.
    This agent is deployed per host and is configured to spawn ``n`` workers.
    When using GPUs, ``n`` maps to the number of GPUs available on the host.

    The local agent does not communicate to other local agents deployed on
    other hosts, even if the workers may communicate inter-host. The worker id
    is interpreted to be a local process. The agent starts and stops all worker
    processes as a single unit.

    The worker function and argument passed to the worker function must be
    python multiprocessing compatible. To pass multiprocessing data structures
    to the workers you may create the data structure in the same multiprocessing
    context as the specified ``start_method`` and pass it as a function argument.

    The exit_barrier_timeout specifies the amount of time (in seconds) to wait
    for other agents to finish. This acts as a safety net to handle cases where
    workers finish at different times, to prevent agents from viewing workers
    that finished early as a scale-down event. It is strongly advised that the
    user code deal with ensuring that workers are terminated in a synchronous
    manner rather than relying on the exit_barrier_timeout.

    Example

    ::

        def trainer(shared_queue):
            pass

        def main():
            start_method="spawn"
            shared_queue= multiprocessing.get_context(start_method).Queue()
            spec = WorkerSpec(
                        role="trainer",
                        local_world_size=nproc_per_process,
                        fn=trainer,
                        args=(shared_queue,),
                        ...<OTHER_PARAMS...>)
            agent = LocalElasticAgent(spec, start_method)
            agent.run()
    """

    def __init__(
        self, spec: WorkerSpec, start_method="spawn", exit_barrier_timeout: float = 300
    ):
        super().__init__(spec, exit_barrier_timeout)
        self._start_method = start_method
        # pyre-ignore[8]: Attribute has type `ProcessContext`; used as `None`.
        self._process_context: mp.ProcessContext = None
        # a queue that holds return values for each worker fn
        # each element of the queue is a tuple (rank, ret_val)
        self._ret_val_queue = None

    @prof
    def _stop_workers(self, worker_group: WorkerGroup) -> None:
        for proc in self._process_context.processes:
            if proc.is_alive():
                proc.terminate()
            proc.join()

    @prof
    def _start_workers(self, worker_group: WorkerGroup) -> Dict[int, Any]:
        spec = worker_group.spec
        store = worker_group.store
        master_addr, master_port = super()._get_master_addr_port(store)
        restart_count = spec.max_restarts - self._remaining_restarts

        dist_infos: Dict[int, _DistInfo] = {}
        for worker in worker_group.workers:
            local_rank = worker.local_rank
            dist_infos[local_rank] = _DistInfo(
                worker.global_rank,
                worker_group.group_rank,
                worker.role_rank,
                spec.local_world_size,
                worker.role_world_size,
                worker.world_size,
                master_addr,
                master_port,
                restart_count,
                spec.max_restarts,
                spec.rdzv_handler.get_run_id(),
                spec.role,
            )

        self._ret_val_queue = mp.get_context(self._start_method).Queue()
        self._process_context = mp.start_processes(
            fn=_wrap,
            args=(self._ret_val_queue, dist_infos, spec.fn, spec.args),
            nprocs=spec.local_world_size,
            join=False,
            daemon=False,
            start_method=self._start_method,
        )

        return {
            local_rank: pid
            for local_rank, pid in enumerate(self._process_context.pids())
        }

    @prof
    def _monitor_workers(self, worker_group: WorkerGroup) -> MonitorResult:
        role = worker_group.spec.role

        # torch process context join() isn't really a join in the
        # traditional sense, it returns True if all the workers have
        # successfully finished, False if some/all are still running
        # and throws an Exception if some/all of them failed
        # passing timeout < 0 means check worker status and return immediately

        worker_pids = {w.id for w in worker_group.workers}
        pc_pids = set(self._process_context.pids())
        if worker_pids != pc_pids:
            log.error(f"[{role}] worker pids do not match process_context pids")
            return MonitorResult(WorkerState.UNKNOWN)

        try:
            if self._process_context.join(timeout=-1):
                # copy ret_val_queue into a map
                ret_vals = {}
                assert worker_group.spec.local_world_size == self._ret_val_queue.qsize()
                for _ in range(self._ret_val_queue.qsize()):
                    (rank, out) = self._ret_val_queue.get()
                    ret_vals[rank] = out
                self._ret_val_queue = None
                return MonitorResult(WorkerState.SUCCEEDED, ret_vals)
            else:
                return MonitorResult(WorkerState.HEALTHY)
        except Exception as e:
            log.exception(f"[{role}] Worker group failed")
            return MonitorResult(
                WorkerState.FAILED,
                exceptions={w.global_rank: e for w in worker_group.workers},
            )



#!/usr/bin/env python3

# Copyright (c) Facebook, Inc. and its affiliates.
# All rights reserved.
#
# This source code is licensed under the BSD-style license found in the
# LICENSE file in the root directory of this source tree.


import torchelastic.tsm.driver.api as Scheduler
import torchelastic.tsm.driver.local_scheduler as local_scheduler


def get_scheduler(scheduler_type: str, **scheduler_params) -> Scheduler:
    schedulers = {"local": local_scheduler.create_scheduler}
    if scheduler_type not in schedulers:
        raise ValueError(f"Unknown scheduler type: {scheduler_type}")
    scheduler_create_method = schedulers[scheduler_type]
    return scheduler_create_method(**scheduler_params)



#!/usr/bin/env python3

# Copyright (c) Facebook, Inc. and its affiliates.
# All rights reserved.
#
# This source code is licensed under the BSD-style license found in the
# LICENSE file in the root directory of this source tree.

import os


def get_env_variable_or_raise(env_name: str) -> str:
    r"""
    Tries to retrieve environment variable. Raises ``ValueError``
    if no environment variable found.

    Arguments:
        env_name (str): Name of the env variable
    """
    value = os.environ.get(env_name, None)
    if value is None:
        msg = f"Environment variable {env_name} expected, but not set"
        raise ValueError(msg)
    return value



#!/usr/bin/env python3

# Copyright (c) Facebook, Inc. and its affiliates.
# All rights reserved.
#
# This source code is licensed under the BSD-style license found in the
# LICENSE file in the root directory of this source tree.


import subprocess
from typing import Optional

import torchelastic.multiprocessing.error_reporter as error_reporter


class ResponsivePopen(subprocess.Popen):
    """
    Wait for child process to terminate. After wait is finished, the
    error reporter will check whether there were any exceptions.
    """

    def __init__(self, non_python=False, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self.non_python = non_python

    def wait_or_raise(self, timeout: Optional[float] = None):
        return_code = super().wait(timeout)
        error_message = error_reporter.get_error(self.pid)

        if error_message:
            raise Exception(error_message)
        return return_code



#!/usr/bin/env python3

# Copyright (c) Facebook, Inc. and its affiliates.
# All rights reserved.
#
# This source code is licensed under the BSD-style license found in the
# LICENSE file in the root directory of this source tree.
# Multiprocessing error-reporting module


from abc import ABC, abstractmethod
from typing import Optional


# Error reported uses files to communicate between processes.
# The name of the environment variable that contains path to the file
# with error message and stacktrace raised by child process.
# Each process has a unique error file.
ERROR_FILE_ENV: str = "ERROR_FILE_ENV_VAR"


class ErrorHandler(ABC):
    @abstractmethod
    def configure(self):
        raise NotImplementedError

    @abstractmethod
    def construct_error_message(self, child_pid: int) -> Optional[str]:
        raise NotImplementedError

    @abstractmethod
    def record_exception(self, e: Exception) -> None:
        raise NotImplementedError


class LocalErrorHandler(ErrorHandler):
    """
    Since we use stderr to get the errors for oss, the local error handler does
    do anything.
    """

    def configure(self):
        pass

    def construct_error_message(self, child_pid: int) -> Optional[str]:
        pass

    def record_exception(self, e: Exception) -> None:
        pass



#!/usr/bin/env/python3

# Copyright (c) Facebook, Inc. and its affiliates.
# All rights reserved.
#
# This source code is licensed under the BSD-style license found in the
# LICENSE file in the root directory of this source tree.

"""Events API

**Overview**:

The event API is used to publish and process events. The module can be used
to capture state transitions to debug or give insight into the execution
flow. The :py:class:`torchelastic.events.Event` is an object containing information
about an occurrence during the execution of the program. The destination handler
for events can be configured by the event type parameter.


.. note:: The event type ``torchelastic`` is reserved by torchelastic for
          platform level events that can be produced by the agent or worker.


**Record Events**:

The event module resembles python's logging framework in terms
of usage and consists of two parts: handler configuration and event publishing.

The example below shows the simple event publishing mechanism
via :py:meth:`torchelastic.events.record_event` method.

::

  from torchelastic.events import Event, record_event

  event.configure(event.NullEventHandler()) # uses event_type = "torchelastic"
  event.configure(event.ConsoleEventHandler(), event_type = "foo")

  def execute():
    event = Event(name="test_event", event_type="foo")
    # The code below will be processed by the ConsoleEventHandler
    record_event(event)

Another way of using the module is via
:py:meth:`torchelastic.events.record`.

::

  from torchelastic.events import record

  def execute():
    metadata = {'key':'value'}
    record(event_name="test", event_type="console", metadata=metadata)


**Writing a Custom Event Handler**:

The custom event handler can be implemented by extending
:py:class:`torchelastic.events.EventHandler` class.

Example

::

  from torchelastic.events import EventHandler

  class StdoutEventHandler(EventHandler):
     def record(self, event):
         print(f"[{event.event_type}]: event_name: {event.name}")

  event.configure(StdoutEventHandler(), event_type="stdout_events")

Now all events with event_type 'stdout_events' will be printed to stdout.

"""

from .api import (  # noqa F401
    ConsoleEventHandler,
    Event,
    EventHandler,
    configure,
    record,
    record_event,
)



#!/usr/bin/env python3

# Copyright (c) Facebook, Inc. and its affiliates.
# All rights reserved.
#
# This source code is licensed under the BSD-style license found in the
# LICENSE file in the root directory of this source tree.
# Multiprocessing error-reporting module

# Note: This is an experimental API and is going to be removed in future
from .api import exec_fn, get_error, record  # noqa F401



#!/usr/bin/env python3

# Copyright (c) Facebook, Inc. and its affiliates.
# All rights reserved.
#
# This source code is licensed under the BSD-style license found in the
# LICENSE file in the root directory of this source tree.

import abc
import os
from dataclasses import dataclass, field
from enum import Enum
from string import Template
from typing import Any, Dict, List, Optional


@dataclass
class Resources:
    """
    Represents resource requirements for a ``Container``.

    Arguments:
            cpu: number of cpu cores (note: not hyper threads)
            gpu: number of gpus
            memMB: MB of ram
            capabilities: additional hardware specs (interpreted by scheduler)
    """

    cpu: int
    gpu: int
    memMB: int
    capabilities: Dict[str, Any] = field(default_factory=dict)

    @staticmethod
    def copy(original: "Resources", **capabilities):
        """
        Copies a resource and applies new capabilities. If the same capabilities
        are present in the original resource and as parameter, the one from parameter
        will be used.
        """
        res_capabilities = dict(original.capabilities)
        res_capabilities.update(capabilities)
        return Resources(
            cpu=original.cpu,
            gpu=original.gpu,
            memMB=original.memMB,
            capabilities=res_capabilities,
        )


# sentinel value used for cases when resource does not matter (e.g. ignored)
NULL_RESOURCE: Resources = Resources(cpu=-1, gpu=-1, memMB=-1)


@dataclass
class Container:
    """
    Represents the specifications of the container that instances of ``Roles``
    run on. Maps to the container abstraction that the underlying scheduler
    supports. This could be an actual container (e.g. Docker) or a physical
    instance depending on the scheduler.

    An ``image`` is a software bundle that is installed on a ``Container``.
    The container on the scheduler dictates what an image actually is.
    An image could be as simple as a tar-ball or map to a docker image.
    The scheduler typically knows how to "pull" the image given an
    image name (str), which could be a simple name (e.g. docker image) or a url
    (e.g. s3://path/my_image.tar).

    Usage:

    ::

     my_container = Container(image="pytorch/torch:1")
                       .require(Resources(cpu=1, gpu=1, memMB=500))
                       .ports(tcp_store=8080, tensorboard=8081)
    """

    image: str
    resources: Resources = NULL_RESOURCE
    port_map: Dict[str, int] = field(default_factory=dict)

    def require(self, resources: Resources) -> "Container":
        """
        Sets resource requirements on the container.
        """
        self.resources = resources
        return self

    def ports(self, **kwargs: int) -> "Container":
        """
        Adds a port mapping for the container
        """
        self.port_map.update({**kwargs})
        return self


# sentinel value used to represent missing string attributes, such as image or entrypoint
MISSING: str = "<MISSING>"

# sentinel value used as the "zero" element in the container group
NULL_CONTAINER: Container = Container(image=MISSING)


class macros:
    """
    Defines macros that can be used with ``Role.entrypoint`` and ``Role.args``.
    The macros will be substituted at runtime to their actual values.

    Available macros:

    1. ``img_root`` - root directory of the pulled image on the container
    2. ``app_id`` - application id (same as the return value of ``session.run(app)``)
    3. ``replica_id`` - unique id for each instance of a replica of a Role,
                        for instance a role with 3 replicas could have the 0, 1, 2
                        as replica ids. Note that when the container fails and is
                        replaced, the new container will have the same ``replica_id``
                        as the one it is replacing. For instance if node 1 failed and
                        was replaced by the scheduler the replacing node will also
                        have ``replica_id=1``.

    Example:

    ::

     # runs: hello_world.py --app_id ${app_id}
     trainer = Role(name="trainer").runs("hello_world.py", "--app_id", macros.app_id)
     app = Application("train_app").of(trainer)
     app_id = session.run(app)


    """

    img_root = "${img_root}"
    app_id = "${app_id}"
    replica_id = "${replica_id}"

    @staticmethod
    def substitute(args: List[str], img_root: str, app_id: str, replica_id: str):
        args_sub = []
        for arg in args:
            sub = Template(arg).safe_substitute(
                img_root=img_root, app_id=app_id, replica_id=replica_id
            )
            args_sub.append(sub)
        return args_sub


@dataclass
class Role:
    """
    A set of nodes that perform a specific duty within the ``Application``.
    Examples:

    1. Distributed data parallel app - made up of a single role (trainer).

    2. App with parameter server - made up of multiple roles (trainer, ps).

    Usage:

    ::

     trainer = Role(name="trainer")
                 .runs("my_trainer.py", "--arg", "foo", ENV_VAR="FOOBAR")
                 .on(container)
                 .replicas(4)

    Arguments:
            name - name of the role
            entrypoint - command (within the container) to invoke the role
            args - commandline arguments to the entrypoint cmd
            env - environment variable mappings
            container - container to run in
            replicas - number of container replicas to run
    """

    name: str
    entrypoint: str
    args: List[str]
    env: Dict[str, str]
    container: Container
    num_replicas: int

    def __init__(self, name: str):
        self.name = name
        self.entrypoint = MISSING
        self.args = []
        self.env = {}
        self.container = NULL_CONTAINER
        self.num_replicas = 1

    def runs(self, entrypoint: str, *args: str, **kwargs: str) -> "Role":
        self.entrypoint = entrypoint
        self.args += [*args]
        self.env.update({**kwargs})
        return self

    def on(self, container: Container) -> "Role":
        self.container = container
        return self

    def replicas(self, replicas: int) -> "Role":
        self.num_replicas = replicas
        return self


@dataclass
class ElasticRole(Role):
    """
    A ``Role`` for which the user provided ``entrypoint`` is executed with the
    torchelastic agent (in the container). Note that the torchelastic agent
    invokes multiple copies of ``entrypoint``.

    For more information about torchelastic see
    `torchelastic quickstart docs <http://pytorch.org/elastic/0.2.0/quickstart.html>`__.

    .. important:: It is the responsibility of the user to ensure that the
                   container's image includes torchelastic. Since TSM has no
                   control over the build process of the image, it cannot
                   automatically include torchelastic in the container's image.

    The following example launches 2 ``replicas`` (nodes) of an elastic ``my_train_script.py``
    that is allowed to scale between 2 to 4 nodes. Each node runs 8 workers which are allowed
    to fail and restart a maximum of 3 times.

    .. warning:: ``replicas`` MUST BE an integer between (inclusive) ``nnodes``. That is,
                   ``ElasticRole("trainer", nnodes="2:4").replicas(5)`` is invalid and will
                   result in undefined behavior.

    ::

     # effectively runs:
     #    python -m torchelastic.distributed.launch
     #        --nproc_per_node 8
     #        --nnodes 2:4
     #        --max_restarts 3
     #        my_train_script.py --script_arg foo --another_arg bar

     elastic_trainer = ElasticRole("trainer", nproc_per_node=8, nnodes="2:4", max_restarts=3)
                        .runs("my_train_script.py", "--script_arg", "foo", "--another_arg", "bar")
                        .on(container)
                        .replicas(2)

    """

    torchelastic_launch_args: List[str]

    def __init__(self, name: str, **launch_kwargs):
        super().__init__(name)
        self.entrypoint = "python"
        self.args += ["-m", "torchelastic.distributed.launch"]
        self.torchelastic_launch_args = []

        launch_kwargs.setdefault("rdzv_backend", "etcd")
        launch_kwargs.setdefault("rdzv_id", macros.app_id)
        launch_kwargs.setdefault("role", name)

        for (arg, val) in launch_kwargs.items():
            if isinstance(val, bool):
                # treat boolean kwarg as a flag
                if val:
                    self.torchelastic_launch_args += [f"--{arg}"]
            else:
                self.torchelastic_launch_args += [f"--{arg}", str(val)]

    def runs(self, entrypoint: str, *args: str, **kwargs: str) -> "ElasticRole":
        if not os.path.isabs(entrypoint) and not entrypoint.startswith(macros.img_root):
            # make entrypoint relative to {img_root} ONLY if it is not an absolute path
            entrypoint = os.path.join(macros.img_root, entrypoint)

        self.args += self.torchelastic_launch_args
        self.args += [entrypoint, *args]
        self.env.update({**kwargs})
        return self


class RunMode(Enum):
    """
    Run mode of the application. Supported modes are:

    1. HEADLESS - Application is allowed to continue running even when the
                  hosting session has terminated. This is the default mode.

    2. MANAGED - Application's lifetime is bound to the hosting session. If
                 the session goes away, so will the app.
    """

    HEADLESS = 0
    MANAGED = 1


@dataclass
class Application:
    """
    Represents a distributed application made up of multiple ``Roles``.
    Contains the necessary information for the driver to submit this
    app to the scheduler.
    """

    name: str
    roles: List[Role] = field(default_factory=list)
    run_mode: RunMode = RunMode.HEADLESS
    is_attached: bool = False

    def of(self, *roles: Role) -> "Application":
        self.roles += [*roles]
        return self


class AppState(Enum):
    """
    State of the application. An application starts from an initial
    ``UNSUBMITTED`` state and moves through ``SUBMITTED``, ``PENDING``,
    ``RUNNING`` states finally reaching a terminal state:
    ``SUCCEEDED``,``FAILED``, ``CANCELLED``.

    If the scheduler supports preemption, the app moves from a ``RUNNING``
    state to ``PENDING`` upon preemption.

    If the user stops the application, then the application state moves
    to ``STOPPED``, then to ``CANCELLED`` when the job is actually cancelled
    by the scheduler.

    1. UNSUBMITTED - app has not been submitted to the scheduler yet
    2. SUBMITTED - app has been successfully submitted to the scheduler
    3. PENDING - app has been submitted to the scheduler pending allocation
    4. RUNNING - app is running
    5. SUCCEEDED - app has successfully completed
    6. FAILED - app has unsuccessfully completed
    7. CANCELLED - app was cancelled before completing
    """

    UNSUBMITTED = 0
    SUBMITTED = 1
    PENDING = 2
    RUNNING = 3
    SUCCEEDED = 4
    FAILED = 5
    CANCELLED = 6


_TERMINAL_STATES = [AppState.SUCCEEDED, AppState.FAILED, AppState.CANCELLED]


def is_terminal(state: AppState) -> bool:
    return state in _TERMINAL_STATES


@dataclass
class AppStatus:
    """
    The runtime status of the ``Application``.
    """

    state: AppState
    num_restarts: int = 0
    msg: str = ""
    ui_url: Optional[str] = None

    def is_terminal(self) -> bool:
        return is_terminal(self.state)


@dataclass
class DescribeAppResponse:
    """
    Response object returned by ``Scheduler.describe(app)`` API. Contains
    the status and description of the application as known by the scheduler.
    For some schedulers implementations this response object has necessary
    and sufficient information to recreate an ``Application`` object in the
    absence of the hosting ``Session``. For these types of schedulers,
    the user can re-``run()`` the attached application. Otherwise the user
    can only call non-creating methods (e.g. ``wait()``, ``status()``, etc).

    Since this class is a data class and contains many member variables we
    keep the usage simple and provide a no-args constructor and chose to
    access the member vars directly rather than provide accessors.
    """

    app_id: str = "<NOT_SET>"
    state: AppState = AppState.UNSUBMITTED
    num_restarts: int = -1
    msg: str = ""
    ui_url: Optional[str] = None
    # TODO T72035216 add other fields that come back from the scheduler's describe


class Scheduler(abc.ABC):
    """
    An interface abstracting functionalities of a scheduler.
    Implementors need only implement those methods annotated with
    ``@abc.abstractmethod``.
    """

    @abc.abstractmethod
    def submit(self, app: Application, mode: RunMode) -> str:
        """
        Submits the application to be run by the scheduler.

        Returns:
            The application id that uniquely identifies the submitted app.
        """
        raise NotImplementedError()

    @abc.abstractmethod
    def describe(self, app_id: str) -> Optional[DescribeAppResponse]:
        """
        Describes the specified application.

        Returns:
            Application description or ``None`` if the app does not exist.
        """
        raise NotImplementedError()

    def exists(self, app_id: str):
        """
        Returns:
            ``True`` if the app exists (was submitted), ``False`` otherwise
        """
        desc = self.describe(app_id)
        return desc is not None

    @abc.abstractmethod
    def _cancel_existing(self, app_id: str) -> None:
        """
        Kills the application. This method will only be called on an
        application that exists.
        """
        raise NotImplementedError()

    def cancel(self, app_id: str) -> None:
        """
        Cancels/kills the application. This method is idempotent within the same
        thread and is safe to call on the same application multiple times.
        However when called from multiple threads/processes on the same app
        the exact semantics of this method depends on the idempotency guarantees
        of the underlying scheduler API.

        .. note:: This method does not block for the application to reach a
                  cancelled state. To ensure that the application reaches a
                  terminal state use the ``wait`` API.
        """
        if self.exists(app_id):
            self._cancel_existing(app_id)
        else:
            # do nothing if the app does not exist
            return


class UnknownAppException(Exception):
    """
    Raised by ``Session`` APIs when either the application does not
    exist or the application is not owned by the session.
    """

    def __init__(self, app_id: str):
        super().__init__(
            f"Unknown app = {app_id}. Did you forget to call session.run() or session.attach()?"
        )


class AppNotReRunnableException(Exception):
    """
    Raised when the application is not re-runnable. That is, one cannot have two run instances
    of the same application. See ``Session.attach()`` for when this is the case.
    """

    def __init__(self, app: Application):
        super().__init__(
            f"App {app.name} is not re-runnable. To re-run attached apps, use a session/scheduler that supports such action."
        )


class Session(abc.ABC):
    """
    Entrypoint and client-facing API for TSM. Has the methods for the user to
    define and act upon ``Applications``. The ``Session`` is stateful and
    represents a logical workspace of the user. It can be backed by a service
    (e.g. TSM server) for persistence or can be standalone with no persistence
    meaning that the ``Session`` lasts only during the duration of the hosting
    process (see the ``attach()`` API for instructions on re-parenting apps
    between sessions).
    """

    def __init__(self, name: str):
        self._name: str = name

    def name(self) -> str:
        """
        Returns:
            The name of this session.
        """
        return self._name

    def run(self, app: Application, mode: RunMode = RunMode.HEADLESS) -> str:
        # input validation
        if not app.roles:
            raise ValueError(
                f"No roles for app: {app.name}. Did you forget to call app.of(roles..)?"
            )

        for role in app.roles:
            if not role.entrypoint:
                raise ValueError(
                    f"No entrypoint for role: {role.name}. Did you forget to call role.runs(entrypoint, args, env)?"
                )
            if role.num_replicas <= 0:
                raise ValueError(
                    f"Non-positive replicas for role: {role.name}. Did you forget to call role.replicas(positive_number)?"
                )
            if role.container == NULL_CONTAINER:
                raise ValueError(
                    f"No container for role: {role.name}. Did you forget to call role.on(container)"
                )
            if role.container.resources == NULL_RESOURCE:
                raise ValueError(
                    f"No resources for container: {role.container.image}. Did you forget to call container.require(resources)"
                )

        return self._run(app, mode)

    @abc.abstractmethod
    def _run(self, app: Application, mode: RunMode = RunMode.HEADLESS) -> str:
        """
        Runs the given application in the specified mode.

        Returns:
            The application id that is used to call other action APIs on the app

        Raises:
            AppNotReRunnableException - if the session/scheduler does not support re-running attached apps
        """
        raise NotImplementedError()

    @abc.abstractmethod
    def status(self, app_id: str) -> Optional[AppStatus]:
        """
        Returns:
            The status of the application, or ``None`` if the app does not exist anymore
            (e.g. was stopped in the past and removed from the scheduler's backend).

        Raises:
            UnknownAppException - if the app was never run or attached on this session
        """
        raise NotImplementedError()

    @abc.abstractmethod
    def wait(self, app_id: str) -> Optional[AppStatus]:
        """
        Block waits (indefinitely) for the application to complete.
        Possible implementation:

        ::

         while(True):
             app_status = status(app)
             if app_status.is_terminal():
                 return
             sleep(10)

        Returns:
            The terminal status of the application, or ``None`` if the app does not exist anymore

        Raises:
            UnknownAppException - if the app was never run or attached on this session
        """
        raise NotImplementedError()

    @abc.abstractmethod
    def list(self) -> Dict[str, Application]:
        """
        Returns the applications that were run with this session mapped by the app_id.
        The persistence of the session is implementation dependent.
        """
        raise NotImplementedError()

    @abc.abstractmethod
    def stop(self, app_id: str) -> None:
        """
        Stops the application, effectively directing the scheduler to cancel
        the job.

        .. note:: This method returns as soon as the cancel request has been
                  submitted to the scheduler. The application will be in a
                  ``RUNNING`` state until the scheduler actually terminates
                  the job. If the scheduler successfully interrupts the job
                  and terminates it the final state will be ``CANCELLED``
                  otherwise it will be ``FAILED``.

        Raises:
            UnknownAppException - if app was never run or attached on this session
        """
        raise NotImplementedError()

    @abc.abstractmethod
    def attach(self, app_id: str) -> Application:
        """
        Attaches the application represented by the provided ``app_id``.
        If the app is already attached to this session, then does nothing
        and returns the existing ``Application`` object.

        Whether all action APIs are supported on the attached application
        is dependent on the specific implementation of the session and scheduler.
        For instance a ``StandaloneSession`` with no persistent backend cannot
        fully recreate the application without the underlying scheduler
        supporting a ``job_definition`` API that stores the job's template.
        Hence in this case, the user will not be able to re-run the attached
        app and can only call action APIs that require only the ``app_id``
        to execute. Please refer to the specific session and scheduler
        implementation documentations to understand the attach behavior,
        assumptions, and consequences.

        Returns:
            The re-created application object

        Raises:
            UnknownAppException - if app was never run or attached on this session
        """
        raise NotImplementedError()



#!/usr/bin/env python3

# Copyright (c) Facebook, Inc. and its affiliates.
# All rights reserved.
#
# This source code is licensed under the BSD-style license found in the
# LICENSE file in the root directory of this source tree.

import os
import re
import sys

from setuptools import find_packages, setup


def get_version():
    # get version string from version.py
    # TODO: ideally the version.py should be generated when setup is run
    version_file = os.path.join(os.path.dirname(__file__), "torchelastic/version.py")
    version_regex = r"__version__ = ['\"]([^'\"]*)['\"]"
    with open(version_file, "r") as f:
        version = re.search(version_regex, f.read(), re.M).group(1)
        return version


if __name__ == "__main__":
    if sys.version_info < (3, 8):
        sys.exit("python >= 3.8 required for torchelastic")

    with open("README.md", encoding="utf8") as f:
        readme = f.read()

    with open("requirements.txt") as f:
        reqs = f.read()

    version = get_version()
    print("-- Building version: " + version)

    setup(
        # Metadata
        name="torchelastic",
        version=version,
        author="PyTorch Elastic Devs",
        author_email="torchelastic@fb.com",
        description="PyTorch Elastic Training",
        long_description=readme,
        long_description_content_type="text/markdown",
        url="https://github.com/pytorch/elastic",
        license="BSD-3",
        keywords=["pytorch", "machine learning", "elastic", "distributed"],
        python_requires=">=3.8",
        install_requires=reqs.strip().split("\n"),
        include_package_data=True,
        packages=find_packages(exclude=("*.test", "aws*", "*.fb")),
        test_suite="torchelastic.test.suites.unittests",
        # PyPI package information.
        classifiers=[
            "Development Status :: 4 - Beta",
            "Intended Audience :: Developers",
            "Intended Audience :: Science/Research",
            "License :: OSI Approved :: BSD License",
            "Programming Language :: Python :: 3",
            "Programming Language :: Python :: 3.8",
            "Topic :: System :: Distributed Computing",
            "Topic :: Scientific/Engineering :: Artificial Intelligence",
        ],
    )



#!/usr/bin/env python3

# Copyright (c) Facebook, Inc. and its affiliates.
# All rights reserved.
#
# This source code is licensed under the BSD-style license found in the
# LICENSE file in the root directory of this source tree.
import datetime
import socket
from contextlib import closing

import torch.distributed as dist
from torchelastic.utils.logging import get_logger


log = get_logger()

_ADDRESS_IN_USE = "Address already in use"
_CONNECT_TIMEOUT = "connect() timed out."
_SOCKET_TIMEOUT = "Socket Timeout"

_MEMBER_CHECKIN = "_tcp_store/num_members"
_LAST_MEMBER_CHECKIN = "_tcp_store/last_member"


def create_c10d_store(
    is_server: bool,
    server_addr: str,
    server_port: int = -1,
    world_size: int = 1,
    timeout: float = (60 * 10),  # 10 min
    retries=3,
):
    if server_port == -1 and world_size > 1:
        raise ValueError(
            f"server_port must be specified when world_size > 1, got server_port={server_port}, world_size={world_size}"
        )

    if server_port != -1:
        log.info(f"sever_port: {server_port}, specified, ignoring retries")

    # only retry when server_port is NOT static
    attempt = retries if server_port == -1 else 1
    while True:
        if server_port != -1:
            port = server_port
        else:
            port = get_free_port()

        log.info(
            f"Creating c10d store on {server_addr}:{port}\n"
            f"  world_size  : {world_size}\n"
            f"  is_server   : {is_server}\n"
            f"  timeout(sec): {timeout}\n"
        )

        try:
            store = dist.TCPStore(
                host_name=server_addr,
                port=port,
                world_size=world_size,
                is_master=is_server,
                timeout=datetime.timedelta(seconds=timeout),
            )
            _check_full_rank(store, world_size)
            log.info("Successfully created c10d store")
            return store
        except RuntimeError as e:
            # this is brittle, but the underlying exception type is not properly pybinded
            # so we parse the error msg for now, interestingly this is how torch itself
            # detects timeouts and port conflicts in their own unittests
            # see - caffe2/torch/testing/_internal/common_utils.py
            # TODO properly map the exceptions in pybind (c10d/init.cpp)
            if str(e) == _CONNECT_TIMEOUT and not is_server:
                raise TimeoutError(
                    f"timed out waiting for tcp store's server: {server_addr}:{port}"
                ) from e
            elif str(e) == _ADDRESS_IN_USE:  # this will only happen on the server
                if attempt < retries:
                    log.warning(
                        f"port: {port} already in use, attempt: [{attempt}/{retries}]"
                    )
                    attempt += 1
                else:
                    raise IOError(
                        f"on {server_addr}, port: {port} already in use"
                    ) from e
            else:
                raise


def _check_full_rank(store, world_size):
    idx = store.add(_MEMBER_CHECKIN, 1)
    if idx == world_size:
        store.set(_LAST_MEMBER_CHECKIN, "<val_ignored>")

    try:
        store.get(_LAST_MEMBER_CHECKIN)
    except RuntimeError as e:
        if str(e) == _SOCKET_TIMEOUT:
            raise TimeoutError(
                f"timed out waiting for all {world_size} members to join"
            ) from e
        else:
            raise


def get_free_port():
    sock = get_socket_with_port()
    with closing(sock):
        return sock.getsockname()[1]


def get_socket_with_port() -> socket.socket:
    """
    Returns a free port on localhost that is "reserved" by binding a temporary
    socket on it. Close the socket before passing the port to the entity
    that requires it. Usage example

    ::

    sock = _get_socket_with_port()
    with closing(sock):
        port = sock.getsockname()[1]
        sock.close()
        # there is still a race-condition that some other process
        # may grab this port before func() runs
        func(port)
    """

    addrs = socket.getaddrinfo(
        host="localhost", port=None, family=socket.AF_UNSPEC, type=socket.SOCK_STREAM
    )
    for addr in addrs:
        family, type, proto, _, _ = addr
        s = socket.socket(family, type, proto)
        try:
            s.bind(("localhost", 0))
            s.listen(0)
            return s
        except OSError as e:
            s.close()
            log.info("Socket creation attempt failed.", exc_info=e)
    raise RuntimeError("Failed to create a socket")



#!/usr/bin/env python3

# Copyright (c) Facebook, Inc. and its affiliates.
# All rights reserved.
#
# This source code is licensed under the BSD-style license found in the
# LICENSE file in the root directory of this source tree.

import torchelastic.rendezvous.etcd_rendezvous as etcd_rdzv
from torchelastic.rendezvous import (
    RendezvousHandler,
    RendezvousHandlerFactory,
    RendezvousParameters,
)


def get_rendezvous_handler(rdzv_params: RendezvousParameters) -> RendezvousHandler:
    factory = RendezvousHandlerFactory()
    factory.register("etcd", etcd_rdzv.create_rdzv_handler)
    return factory.create_rdzv_handler(rdzv_params)



#!/usr/bin/env python3

# Copyright (c) Facebook, Inc. and its affiliates.
# All rights reserved.
#
# This source code is licensed under the BSD-style license found in the
# LICENSE file in the root directory of this source tree.

import time
from typing import Dict, Optional

from torchelastic.tsm.driver.api import (
    Application,
    AppNotReRunnableException,
    AppStatus,
    RunMode,
    Scheduler,
    Session,
    UnknownAppException,
)


class StandaloneSession(Session):
    def __init__(self, name: str, scheduler: Scheduler, wait_interval: int = 10):
        super().__init__(name)
        self._scheduler = scheduler
        self._wait_interval = wait_interval

        # TODO T72035686 implement an LRU cache (see local_scheduler.py) and use it here and also in local_scheduler
        # app_id -> Application
        self._apps: Dict[str, Application] = {}

    def run(self, app: Application, mode: RunMode = RunMode.HEADLESS) -> str:
        if app.is_attached:
            raise AppNotReRunnableException(app)
        return super().run(app, mode)

    def _run(self, app: Application, mode: RunMode = RunMode.HEADLESS) -> str:
        app_id = self._scheduler.submit(app, mode)
        self._apps[app_id] = app
        return app_id

    def status(self, app_id: str) -> Optional[AppStatus]:
        if app_id not in self._apps:
            raise UnknownAppException(app_id)

        desc = self._scheduler.describe(app_id)
        if not desc:
            # app does not exist on the scheduler; remove it from apps cache
            # effectively removes this app from the list() API
            del self._apps[app_id]
            return None

        app_status = AppStatus(desc.state, desc.num_restarts, desc.msg)
        app_status.ui_url = desc.ui_url
        return app_status

    def wait(self, app_id: str) -> Optional[AppStatus]:
        while True:
            app_status = self.status(app_id)

            if not app_status:
                return None
            if app_status.is_terminal():
                return app_status
            else:
                time.sleep(self._wait_interval)

    def list(self) -> Dict[str, Application]:
        # opportunistically query for each app's status to update the app
        # copy the keys (app ids) since status(app_id) mutates self._apps
        app_ids = list(self._apps.keys())
        for app_id in app_ids:
            self.status(app_id)
        return self._apps

    def stop(self, app_id: str) -> None:
        status = self.status(app_id)
        if status is None or status.is_terminal():
            return  # do nothing; app does not exist or has already finished
        else:
            self._scheduler.cancel(app_id)

    def attach(self, app_id: str) -> Application:
        # TODO recreate the app to the best extent; empty app for now (e.g. cannot re-run the app)
        app = Application(name=app_id)
        app.is_attached = True
        self._apps[app_id] = app
        return app



#!/usr/bin/env python3

# Copyright (c) Facebook, Inc. and its affiliates.
# All rights reserved.
#
# This source code is licensed under the BSD-style license found in the
# LICENSE file in the root directory of this source tree.
# Multiprocessing error-reporting module


from torchelastic.multiprocessing.error_reporter.error_handler import (
    ErrorHandler,
    LocalErrorHandler,
)


def get_error_handler() -> ErrorHandler:
    return LocalErrorHandler()



#!/usr/bin/env/python3

# Copyright (c) Facebook, Inc. and its affiliates.
# All rights reserved.
#
# This source code is licensed under the BSD-style license found in the
# LICENSE file in the root directory of this source tree.

""" Rendezvous

In the context of torchelastic we use the term ``rendezvous`` to refer to
a particular functionality that combines a **distributed
synchronization** primitive with **peer discovery**.

It is used by torchelastic to gather participants of a training job
(i.e. workers) such that they all agree on the same list of participants
and everyone’s roles, as well as make a consistent collective decision
on when training can begin/resume.

Torchelastic Rendezvous provides the following critical functionalities:

**Barrier**:

Workers performing rendezvous will all block until the rendezvous is
considered complete - this happens when at least ``min`` total number of
workers have joined the rendezvous barrier (for the same job). This also
implies the barrier is not necessarily of fixed size.

There’s an additional small waiting time after reaching ``min`` number
of workers - this is used to ensure the rendezvous is not completed “too
quickly” (which could potentially exclude additional workers attempting
to join at approximately the same time).

If ``max`` number of workers is gathered at the barrier, the rendezvous
is completed immediately.

There’s also an overall timeout which causes the rendezvous to fail if
``min`` number of workers is never reached – this is meant to be a
simple fail-safe to help release partially allocated job resources, in
case there’s a problem with the resource manger, and is meant to be
interpreted as non-retryable.

**Exclusivity**:

A simple distributed barrier would not be sufficient, as we also need to
ensure that only one group of workers exists at any given time (for a
given job). In other words, new workers (i.e. joining late) should not
be able to form a parallel independent group of workers for the same
job.

Torchelastic rendezvous ensures that if a group of workers has already
completed a rendezvous (and hence might already be training), then
additional “late” workers attempting to rendezvous will only announce
themselves as waiting, and will have to wait until the (previously
completed) existing rendezvous is destroyed first.

**Consistency**:


When a rendezvous is completed, all its members will agree on the job
membership and everyone’s role in it. This role is represented using an
integer, called rank, that is between between 0 and world size.

Note that ranks are *not stable*, in the sense that the same worker
process can be assigned a different rank in the next (re-)rendezvous.

**Fault-tolerance**:

Torchelastic rendezvous is designed to tolerate worker failures during
the rendezvous process. Should a process crash (or lose network
connectivity, etc), between joining the rendezvous and it being
completed, then a re-rendezvous with remaining healthy workers will
happen automatically.

A worker can also fail *after* it has completed (or *has been
observered* by other workers to have completed) the rendezvous - this
scenario will be handled by the torchelastic ``train_loop`` instead
(where it will also trigger a re-rendezvous).

**Shared key-value store**:

When the rendezvous is completed, a shared key-value store is created
and returned. This store implements a ``torch.distributed.Store`` API
(see `distributed communication
docs <https://pytorch.org/docs/stable/distributed.html>`__).

This store is only shared by the members of the completed rendezvous. It
is intended to be used by torchelastic to exchange information necessary
to initialize job control and data-planes.

**Waiting workers and rendezvous closing**:

Torchelastic rendezvous handler object provides additional
functionalities, which are technically not part of the rendezvous
process:

1. Querying how many workers arrived late at the barrier, who
   can participate in *next* rendezvous.

2. Setting the rendezvous *closed* to signal all workers not
   to participate in next rendezvous.
"""

from .api import (  # noqa: F401
    RendezvousClosedException,
    RendezvousHandler,
    RendezvousHandlerFactory,
    RendezvousNonRetryableError,
    RendezvousParameters,
    RendezvousTimeoutException,
)


try:
    import torchelastic.rendezvous.static_init  # noqa: F401
except ModuleNotFoundError:
    pass



#!/usr/bin/env python3

# Copyright (c) Facebook, Inc. and its affiliates.
# All rights reserved.
#
# This source code is licensed under the BSD-style license found in the
# LICENSE file in the root directory of this source tree.

r"""
Below is a simple program that locally launches a multi-role
(trainer, parameter server, reader) distributed application.
Each ``Role`` runs multiple replicas. In reality each replica
runs on its own container on a host. An ``Application`` is made up
to one or more such ``Roles``.

.. code:: python

 import getpass
 import torchelastic.tsm.driver as tsm

 username = getpass.getuser()
 train_project_dir = tsm.Container(image=f"/home/{username}/pytorch_trainer")
 reader_project_dir = tsm.Container(image=f"/home/{username}/pytorch_reader")

 trainer = tsm.Role(name="trainer")
              .runs("train_main.py", "--epochs", "50", MY_ENV_VAR="foobar")
              .on(train_project_dir)
              .replicas(4)

 ps = tsm.Role(name="parameter_server")
         .run("ps_main.py")
         .on(train_project_dir)
         .replicas(10)

 reader = tsm.Role(name="reader")
             .runs("reader/reader_main.py", "--buffer_size", "1024")
             .on(reader_project_dir)
             .replicas(1)

 app = tsm.Application(name="my_train_job").of(trainer, ps, reader)

 session = tsm.session(name="my_session", scheduler_backend="local")
 app_id = session.run(app)
 session.wait(app_id)

In the example above, we have done a few things:

1. Created and ran a distributed training application that runs a total of
   4 + 10 + 1 = 15 containers (just processes since we used a ``local`` scheduler).
2. The ``trainer`` and ``ps`` run from the same container:
   ``/home/$USER/pytorch_trainer`` and the reader runs from a container:
   ``/home/$USER/pytorch_reader``. The containers map to a local directory
   because we are using a local scheduler. For other non-trivial schedulers
   a container could map to a Docker image, tarball, rpm, etc.
3. The main entrypoints are relative to the container image's root dir.
   For example, the trainer runs ``/home/$USER/pytorch_trainer/train_main.py``.
4. Arguments to each role entrypoint are passed as ``*args`` after the entrypoint CMD.
5. Environment variables to each role entrypoint are passed as ``**kwargs``
   after the arguments.
6. The ``session`` object has action APIs on the app: ``run``, ``describe``, and ``wait``.


"""
from torchelastic.tsm.driver.api import (  # noqa: F401 F403
    Application,
    AppState,
    Container,
    ElasticRole,
    Resources,
    Role,
    RunMode,
    Session,
    macros,
)
from torchelastic.tsm.driver.schedulers import get_scheduler
from torchelastic.tsm.driver.standalone_session import StandaloneSession


try:
    from torchelastic.tsm.driver.api_extended import *  # noqa: F401 F403
except ModuleNotFoundError:
    pass


def session(
    name: str, scheduler_backend: str, backend: str = "standalone", **scheduler_args
):
    scheduler = get_scheduler(scheduler_backend, **scheduler_args)
    if backend != "standalone":
        raise ValueError(
            f"Unsupported session backend: {backend}. Supported values: standalone"
        )
    return StandaloneSession(name=name, scheduler=scheduler)
